from collections import Counter
from typing import Callable, Dict, List, Set, cast
from unittest import mock
from unittest.mock import patch

import pytest
from sqlalchemy.orm import Session
from sqlalchemy.sql.expression import cast as sa_cast

from redun import Handle, Scheduler, catch, cond, task
from redun.backends.db import JSON, REDUN_DB_VERSIONS, CallNode, DBVersionInfo, Execution
from redun.backends.db import Handle as DbHandle
from redun.backends.db import (
    HandleEdge,
    Job,
    PreviewValue,
    RedunBackendDb,
    RedunDatabaseError,
    RedunMigration,
    RedunVersion,
    RedunVersionError,
    Tag,
    TagEntity,
    Value,
    parse_db_version,
)
from redun.config import Config, create_config_section
from redun.file import Dir, File
from redun.functools import seq
from redun.tests.utils import use_tempdir
from redun.utils import PreviewClass


def test_cache_db() -> None:
    backend = RedunBackendDb(db_uri="sqlite:///:memory:")
    backend.load()

    # Set values into cache and retrieve them.
    value_hash1 = backend.record_value("myvalue")
    value_hash2 = backend.record_value(10)
    assert backend.get_value(value_hash1) == ("myvalue", True)
    assert backend.get_value(value_hash2) == (10, True)

    # Unknown value should return None.
    assert backend.get_value("unknown_call_hash") == (None, False)

    # Reinserting should be a no-op.
    backend.record_value("myvalue") == value_hash1


def test_max_value_size() -> None:
    """
    Test max_value_size config variable changes the max size limit for caching.
    """
    config_dict = {"max_value_size": 100}
    config = create_config_section(config_dict)

    backend = RedunBackendDb(db_uri="sqlite:///:memory:", config=config)
    backend.load()
    small_object = "J" * 10000

    with pytest.raises(RedunDatabaseError):
        backend.record_value(small_object)


def test_call_graph_db() -> None:
    backend = RedunBackendDb(db_uri="sqlite:///:memory:")
    backend.load()
    task_hashes = {
        "task_hash1",
        "task_hash2",
        "task_hash3",
    }

    # Build graph.
    call_hash1 = backend.record_call_node(
        task_name="mytask1",
        task_hash="task_hash1",
        args_hash="args_hash1",
        expr_args=((), {}),
        eval_args=((), {}),
        result_hash="result_hash1",
        child_call_hashes=[],
    )
    call_hash2 = backend.record_call_node(
        task_name="mytask2",
        task_hash="task_hash2",
        args_hash="args_hash2",
        expr_args=((), {}),
        eval_args=((), {}),
        result_hash="result_hash2",
        child_call_hashes=[],
    )
    backend.record_call_node(
        task_name="mytask3",
        task_hash="task_hash3",
        args_hash="args_hash3",
        expr_args=((), {}),
        eval_args=((), {}),
        result_hash="result_hash3",
        child_call_hashes=[call_hash1, call_hash2],
    )

    # Get a node.
    call_node = backend._get_call_node("task_hash3", "args_hash3", task_hashes)
    assert call_node
    assert call_node.task_name == "mytask3"

    # Ensure we can fetch the children.
    assert call_node.children[0].task_name == "mytask1"
    assert call_node.children[1].task_name == "mytask2"

    # Get another node.
    call_node = backend._get_call_node("task_hash1", "args_hash1", task_hashes)
    assert call_node
    assert call_node.task_name == "mytask1"

    # Ensure we can get the parents.
    call_node.parents[0].task_name == "mytask3"


def test_handle_db() -> None:
    backend = RedunBackendDb(db_uri="sqlite:///:memory:")
    backend.load()

    # Create a Handle and pass it through three calls.
    handle = Handle("conn")
    handle2 = handle.apply_call("call_hash2")
    handle3 = handle.apply_call("call_hash3")
    handle4 = handle3.apply_call("call_hash4")
    backend.advance_handle([handle], handle2)
    backend.advance_handle([handle], handle3)
    backend.advance_handle([handle3], handle4)

    # We should have 4 handles and 3 edges.
    assert backend.session
    assert backend.session.query(DbHandle).count() == 4
    assert backend.session.query(HandleEdge).count() == 3

    # They should all be valid.
    assert backend.is_valid_handle(handle)
    assert backend.is_valid_handle(handle2)
    assert backend.is_valid_handle(handle3)
    assert backend.is_valid_handle(handle4)

    # Let's rollback to first handle.
    backend.rollback_handle(handle)

    # Descendants should be invalid.
    assert backend.is_valid_handle(handle)
    assert not backend.is_valid_handle(handle2)
    assert not backend.is_valid_handle(handle3)
    assert not backend.is_valid_handle(handle4)


def test_scheduler_db():
    """
    Scheduler should be able to use the database Backend.
    """
    backend = RedunBackendDb(db_uri="sqlite:///:memory:")
    backend.load()
    scheduler = Scheduler(backend=backend)

    @task()
    def my_task1():
        return 10

    @task()
    def my_task2():
        return my_task1()

    result = scheduler.run(my_task2())
    assert result == 10

    result = scheduler.run(my_task2())
    assert result == 10


def test_db_migrate() -> None:
    """
    Ensure we can safely detect database versions.
    """
    backend = RedunBackendDb(db_uri="sqlite:///:memory:")
    backend.load()
    assert backend.session

    # Assert fresh database is current version and is compatible.
    version = backend.get_db_version()
    latest_version = backend.get_all_db_versions()[-1]
    assert version == latest_version
    assert backend.is_db_compatible()

    # Make an older database (one with a version incompatible with the latest one)
    min_compat_for_latest_version, _ = backend.get_db_version_required()
    prev_incompat_version = [
        v for v in backend.get_all_db_versions() if v < min_compat_for_latest_version
    ][-1]
    backend.migrate(prev_incompat_version)

    # Database should no longer be compatible with library.
    assert not backend.is_db_compatible()

    # Migrating should update the database version.
    backend.migrate()
    assert backend.is_db_compatible()
    assert backend.get_db_version() == latest_version

    # Migrate should not downgrade when upgrade_only=True.
    backend.migrate(desired_version=parse_db_version("1.0"), upgrade_only=True)
    assert backend.get_db_version() == latest_version

    # Make a newer database.
    version_row = (
        backend.session.query(RedunVersion).order_by(RedunVersion.timestamp.desc()).first()
    )
    version_row.version = latest_version.major + 1
    backend.session.add(version_row)
    backend.session.commit()

    with pytest.raises(RedunVersionError, match="too new"):
        backend.migrate()


def test_db_migrate_minor() -> None:
    """
    Ensure we can support minor version numbers.
    """
    backend = RedunBackendDb(db_uri="sqlite:///:memory:")
    backend.load()
    assert backend.session

    latest_version = REDUN_DB_VERSIONS[-1]
    new_migration_id = "abcd1234abcd"
    new_version = DBVersionInfo(
        new_migration_id,
        latest_version.major,
        latest_version.minor + 1,
        "Example minor version.",
    )
    versions_mock = REDUN_DB_VERSIONS + [new_version]

    def upgrade(config, migration_id):
        # Simulate alembic upgrade.
        backend.session.add(RedunMigration(version_num=migration_id))
        backend.session.commit()

    with patch("redun.backends.db.REDUN_DB_VERSIONS", versions_mock), patch(
        "redun.backends.db.upgrade", side_effect=upgrade
    ) as upgrade_mock:
        backend.migrate(new_version)

        # Alembic upgrade should be called.
        assert upgrade_mock.call_args[0][1] == new_migration_id

        # Version should now be updated.
        assert backend.get_db_version() == new_version


def test_task_args(scheduler: Scheduler, session: Session) -> None:
    """
    Arguments to a task should be recorded.
    """

    @task()
    def task1(a, b, c=10, d=20):
        return a

    assert scheduler.run(task1(1, 2, c=3)) == 1

    last_exec = session.query(Execution).one()
    [arg1, arg2, arg3, arg4] = sorted(
        last_exec.call_node.arguments, key=lambda arg: arg.arg_key or str(arg.arg_position)
    )

    # Positional arguments.
    assert arg1.call_hash == last_exec.call_node.call_hash
    assert arg1.arg_position == 0
    assert arg1.arg_key is None
    assert arg1.value_parsed == 1

    assert arg2.call_hash == last_exec.call_node.call_hash
    assert arg2.arg_position == 1
    assert arg2.arg_key is None
    assert arg2.value_parsed == 2

    # Keyword argument.
    assert arg3.call_hash == last_exec.call_node.call_hash
    assert arg3.arg_position is None
    assert arg3.arg_key == "c"
    assert arg3.value_parsed == 3

    # Default argument.
    assert arg4.call_hash == last_exec.call_node.call_hash
    assert arg4.arg_position is None
    assert arg4.arg_key == "d"
    assert arg4.value_parsed == 20


def test_multiple_expr(scheduler: Scheduler) -> None:
    """
    Multiple references of the same expression should only run the task once.
    """
    task_calls = []

    @task()
    def task1():
        task_calls.append("task1")
        return 1

    @task()
    def task2(y):
        return [y, task1()]

    @task()
    def main(x):
        return [task2(1), task2(2)]

    assert scheduler.run(main(1)) == [[1, 1], [2, 1]]
    assert task_calls == ["task1"]

    # Now let's test the cache.
    with mock.patch.object(scheduler, "_get_cache", wraps=scheduler._get_cache) as get_cache:
        assert scheduler.run(main(1)) == [[1, 1], [2, 1]]
        # Depending on timing, we'll get a CSE from pending or the backend.
        assert 4 <= get_cache.call_count <= 5

    # Get root call node.
    backend = scheduler.backend
    assert isinstance(backend, RedunBackendDb) and backend.session

    call_node = backend.session.query(CallNode).filter_by(task_name="main").one()

    # Assert diamond call graph.
    assert len(call_node.children) == 2
    assert (
        call_node.children[0].children[0].call_hash == call_node.children[1].children[0].call_hash
    )


def test_parent_job(scheduler: Scheduler, session: Session) -> None:
    """
    Parent job should be set for subtasks and arguments.

    By setting the parent job we can trace where an argument came from, i.e.
    the upstream CallNode.
    """

    @task()
    def task1():
        return 10

    @task()
    def task2(x):
        return x

    @task()
    def task3():
        return task2(task1())

    assert scheduler.run(task3()) == 10

    last_exec = session.query(Execution).one()
    assert last_exec.job.task.name == "task3"
    assert last_exec.job.child_jobs[0].task.name == "task1"
    assert last_exec.job.child_jobs[1].task.name == "task2"
    assert last_exec.job.child_jobs[1].call_node.arguments[0].upstream[0].task_name == "task1"


def test_parent_job_simple(scheduler: Scheduler, session: Session) -> None:
    """
    Parent job should be set for subtasks even through SimpleExpression.
    """

    @task()
    def task1():
        return {"a": 10}

    @task()
    def task2():
        return task1()["a"]

    assert scheduler.run(task2()) == 10

    last_exec = session.query(Execution).one()
    assert last_exec.job.task.name == "task2"
    assert last_exec.job.child_jobs[0].task.name == "task1"


def test_parent_job_simple_arg(scheduler: Scheduler, session: Session) -> None:
    """
    Parent job should be set for subtasks even through SimpleExpression as arguments.

    The upstream CallNode should also be recorded for such arguments.
    """

    @task()
    def task1():
        return {"a": 10}

    @task()
    def task2(x):
        return x

    @task()
    def task3():
        return task2(task1()["a"])

    assert scheduler.run(task3()) == 10

    last_exec = session.query(Execution).one()
    assert last_exec.job.task.name == "task3"
    assert last_exec.job.child_jobs[0].task.name == "task1"
    assert last_exec.job.child_jobs[1].task.name == "task2"
    assert last_exec.job.child_jobs[1].call_node.arguments[0].upstream[0].task_name == "task1"


def test_parent_job_multiple_simple_arg(scheduler: Scheduler, session: Session) -> None:
    """
    Parent job should be set for subtasks even through SimpleExpression as arguments.
    """

    @task()
    def task1():
        return {"a": 10}

    @task()
    def task2(x, y):
        return x

    @task()
    def task3():
        return task2(task1()["a"], task1()["a"])

    assert scheduler.run(task3()) == 10

    last_exec = session.query(Execution).one()
    assert last_exec.job.task.name == "task3"

    # Due to collapsing multiple equivalent expressions, there should only be
    # two sub jobs.
    assert len(last_exec.job.child_jobs) == 2
    assert last_exec.job.child_jobs[0].task.name == "task1"
    assert last_exec.job.child_jobs[1].task.name == "task2"

    # Both arguments should be recorded and both of their upstreams should be set.
    assert len(last_exec.job.child_jobs[1].call_node.arguments) == 2
    assert last_exec.job.child_jobs[1].call_node.arguments[0].upstream[0].task_name == "task1"
    assert last_exec.job.child_jobs[1].call_node.arguments[1].upstream[0].task_name == "task1"


def test_parent_job_simple_call(scheduler: Scheduler, session: Session) -> None:
    """
    Parent job should be set for subtask that were returned as result values.
    """

    @task()
    def task1():
        return 10

    @task()
    def task2():
        return task1

    @task()
    def task3():
        return task2()()

    assert scheduler.run(task3()) == 10

    last_exec = session.query(Execution).one()
    assert last_exec.job.task.name == "task3"
    assert last_exec.job.child_jobs[0].task.name == "task2"
    assert last_exec.job.child_jobs[1].task.name == "task1"


def test_task_value_recording(scheduler: Scheduler, session: Session) -> None:
    """
    For every Task recorded in the database, a corresponding Value record should exist. This
    should be true even for Tasks that aren't passed as arguments to other Tasks.
    """
    from redun.backends.db import Task as DBTask
    from redun.backends.db import Value as DBValue

    @task()
    def task1(a: int, b: int) -> int:
        return a + b

    scheduler.run(task1(1, 2))
    assert (
        session.query(DBValue).join(DBTask, DBTask.hash == DBValue.value_hash).one().value_hash
        == task1.hash
    )

    @task()
    def main(tsk: Callable[[int, int], int]):
        return [tsk(1, 10), tsk(1, 12)]

    scheduler.run(main(task1))

    assert session.query(DBValue).join(DBTask, DBTask.hash == DBValue.value_hash).count() == 2


@use_tempdir
def test_subvalues(scheduler: Scheduler, session: Session) -> None:
    """
    Subvalues should be recorded.
    """
    file1 = File("file1")
    file1.write("hello")

    file2 = File("file2")
    file2.write("goodbye")

    @task()
    def main(files: List[File]) -> Dict[File, str]:
        return {file: cast(str, file.read()) for file in files}

    assert scheduler.run(main([file1, file2])) == {file1: "hello", file2: "goodbye"}

    last_exec = session.query(Execution).one()

    # Subvalues (files) of the List[File] argument should be recorded.
    subvalues = last_exec.call_node.arguments[0].value.children
    paths = {subvalue.value_parsed.path for subvalue in subvalues}
    assert paths == {"file1", "file2"}

    # Subvalues (files) of the return value Dict[File, str] should be recorded.
    subvalues2 = last_exec.call_node.value.children
    paths2 = {subvalue.value_parsed.path for subvalue in subvalues2}
    assert paths2 == {"file1", "file2"}


@use_tempdir
def test_dir_subvalues(scheduler: Scheduler, session: Session) -> None:
    """
    Subvalues of a Dir should be recorded.
    """
    file1 = File("dir/file1")
    file1.write("hello")

    file2 = File("dir/file2")
    file2.write("goodbye")

    dir = Dir("dir")

    @task()
    def main(dir: Dir) -> Set[str]:
        return {cast(str, file.read()) for file in dir}

    assert scheduler.run(main(dir)) == {"hello", "goodbye"}

    last_exec = session.query(Execution).one()

    # Dir should be recorded as argument.
    arg_value = last_exec.call_node.arguments[0].value
    dir2 = arg_value.value_parsed
    assert isinstance(dir2, Dir)
    assert dir2.path == dir.path

    # Subvalues (files) of the Dir argument should be recorded.
    subvalues = arg_value.children
    paths = {subvalue.value_parsed.path for subvalue in subvalues}
    assert paths == {"dir/file1", "dir/file2"}


@use_tempdir
def test_nested_dir_subvalues(scheduler: Scheduler, session: Session) -> None:
    """
    Subvalues of a Dir nested inside another Value should be recorded.
    """
    file1 = File("dir1/file1")
    file1.write("hello")

    file2 = File("dir2/file2")
    file2.write("goodbye")

    file3 = File("dir2/file3")
    file3.write("world")

    dir1 = Dir("dir1")
    dir2 = Dir("dir2")

    @task()
    def main(dirs: List[Dir]) -> Set[str]:
        return {cast(str, file.read()) for dir in dirs for file in dir}

    assert scheduler.run(main([dir1, dir2])) == {"hello", "goodbye", "world"}

    last_exec = session.query(Execution).one()

    # List of Dirs should be recorded.
    arg_value = last_exec.call_node.arguments[0].value
    dir_list = arg_value.value_parsed
    assert isinstance(dir_list, list)
    assert [dir.path for dir in dir_list] == ["dir1", "dir2"]

    # All subvalues, Dirs and Files, should be recorded.
    paths = {child.value_parsed.path for child in arg_value.children}
    assert paths == {"dir1", "dir1/file1", "dir2", "dir2/file2", "dir2/file3"}


def test_execution_job(scheduler: Scheduler, session: Session) -> None:
    """
    We should be able to query an Execution's job tree.
    """

    @task()
    def task1(x):
        return x + 1

    @task()
    def main():
        return [task1(1), task1(2)]

    assert scheduler.run(main()) == [2, 3]

    # We should be able to query the job tree.
    last_exec = session.query(Execution).one()
    assert len(last_exec.jobs) == 3
    assert {job.task.name for job in last_exec.jobs} == {"task1", "main"}


def test_job_status(scheduler: Scheduler, session: Session) -> None:
    """
    We should be able to classify a Job's status.
    """

    @task()
    def task1(x):
        assert x != 0
        return x + 1

    @task()
    def main():
        return [task1(1), task1(0)]

    with pytest.raises(AssertionError):
        scheduler.run(main())

    with pytest.raises(AssertionError):
        scheduler.run(main())

    [exec1, exec2] = (
        session.query(Execution).join(Job, Job.id == Execution.job_id).order_by(Job.start_time)
    )

    assert Counter(job.status for job in exec1.jobs) == Counter({"FAILED": 2, "DONE": 1})
    assert Counter(job.status for job in exec2.jobs) == Counter({"FAILED": 2, "CACHED": 1})


def test_catch_downstream(scheduler: Scheduler) -> None:
    """
    Result from catch should properly propagate to downstream tasks.
    """

    @task()
    def task1(x):
        return x + 1

    @task()
    def recover(error):
        return error

    @task()
    def main():
        x = catch(task1(1), ValueError, recover)
        return task1(x)

    assert scheduler.run(main()) == 3


def test_cond_downstream(scheduler: Scheduler, session: Session) -> None:
    @task()
    def id(x):
        return x

    @task()
    def main(pred=True):
        x = cond(id(pred), id(2), id(3))
        return id(x)

    assert scheduler.run(main()) == 2


def test_catch_dataflow(scheduler: Scheduler, session: Session) -> None:
    """
    SchedulerTasks with partially evaluated args should have data provenance.
    """

    @task()
    def task1(x):
        if x:
            raise ValueError("BOOM")
        return x

    @task()
    def task2(x):
        return x

    @task()
    def recover(error):
        return "recover"

    @task()
    def main():
        x = catch(task1(True), ValueError, recover)
        return task2(x)

    @task()
    def main2():
        x = catch(task1(False), ValueError, recover)
        return task2(x)

    assert scheduler.run(main()) == "recover"
    assert scheduler.run(main2()) is False

    [exec1, exec2] = (
        session.query(Execution).join(Job, Job.id == Execution.job_id).order_by(Job.start_time)
    )

    # Assert dataflow for exec1.
    [task2_node, task1_node, recover_node] = exec1.call_node.children
    assert task2_node.task.name == "task2"
    assert task1_node.task.name == "task1"
    assert recover_node.task.name == "recover"
    assert task2_node.arguments[0].upstream == [recover_node]
    assert recover_node.arguments[0].upstream == [task1_node]

    # Assert dataflow for exec2.
    [task2_node, task1_node] = exec2.call_node.children
    assert task2_node.task.name == "task2"
    assert task1_node.task.name == "task1"
    assert task2_node.arguments[0].upstream == [task1_node]


def test_catch_reraise_dataflow(scheduler: Scheduler, session: Session) -> None:
    """
    catch with a reraise should have proper dataflow.
    """

    @task()
    def task1():
        raise ValueError("BOOM")

    @task()
    def recover(error):
        raise error

    @task()
    def main():
        return catch(task1(), ValueError, recover)

    with pytest.raises(ValueError):
        scheduler.run(main())

    exec1 = session.query(Execution).one()
    assert exec1.job.child_jobs[1].call_node.task.name == "recover"
    assert isinstance(exec1.job.child_jobs[1].call_node.value.value_parsed.error, ValueError)


def test_seq_dataflow(scheduler: Scheduler, session: Session) -> None:
    """
    seq() should record a proper dataflow.
    """

    @task()
    def inc(x: int) -> int:
        return x + 1

    @task()
    def main():
        return seq([inc(1), inc(2), inc(3)])

    assert scheduler.run(main()) == [2, 3, 4]

    exec1 = session.query(Execution).one()
    assert [job.call_node.value.value_parsed for job in exec1.job.child_jobs] == [2, 3, 4]


def test_cond_dataflow(scheduler: Scheduler, session: Session) -> None:
    @task()
    def pred(x):
        return x

    @task()
    def then(x):
        return x

    @task()
    def else_(x):
        return x

    @task()
    def follow_up(x):
        return x

    @task()
    def main(arg=True):
        x = cond(pred(arg), then(2), else_(3))
        return follow_up(x)

    assert scheduler.run(main(True)) == 2
    assert scheduler.run(main(False)) == 3

    [exec1, exec2] = (
        session.query(Execution).join(Job, Job.id == Execution.job_id).order_by(Job.start_time)
    )

    # Assert dataflow for exec1.
    [follow_up_node, pred_node, then_node] = exec1.call_node.children
    [follow_up_node, pred_node, then_node] = exec1.call_node.children
    assert follow_up_node.task.name == "follow_up"
    assert pred_node.task.name == "pred"
    assert then_node.task.name == "then"
    assert set(follow_up_node.arguments[0].upstream) == {pred_node, then_node}

    # Assert dataflow for exec2.
    [follow_up_node, pred_node, else_node] = exec2.call_node.children
    assert follow_up_node.task.name == "follow_up"
    assert pred_node.task.name == "pred"
    assert else_node.task.name == "else_"
    assert set(follow_up_node.arguments[0].upstream) == {pred_node, else_node}


def test_simple_expression_upstreams(scheduler: Scheduler, session: Session) -> None:
    """
    We should be able to record the upstream of a cached SimpleExpression.
    """

    @task()
    def get_data() -> dict:
        return {"key": 10}

    @task()
    def add(x: int, y: int) -> int:
        return x + y

    @task()
    def main(a: int) -> int:
        data = get_data()
        x = data["key"]
        z = add(x, a)
        return z

    assert scheduler.run(main(1)) == 11

    # Redefine the task.
    @task()  # type: ignore[no-redef]
    def get_data() -> dict:
        return {"key": -10}

    # Running the scheduler a second time will use the cached result for main(), which is
    #   TaskExpression('add',
    #     (SimpleExpression('getitem', (
    #         TaskExpression('get_data', (), {}),
    #         'key',
    #       ), {})),
    #     {}
    #   )
    #
    # The SimpleExpression will deserialized from cache and will need its upstreams
    # set to TaskExpression('get_data', (), {})
    assert scheduler.run(main(1)) == -9

    def get_args(call_node):
        return sorted(call_node.arguments, key=lambda arg: arg.arg_position)

    execs = session.query(Execution).all()
    assert get_args(execs[0].call_node.children[0])[0].upstream[0].task_name == "get_data"
    assert get_args(execs[1].call_node.children[0])[0].upstream[0].task_name == "get_data"


@use_tempdir
def test_db_automigrate() -> None:
    """
    Backend should respect automigration config option.
    """
    backend = RedunBackendDb(db_uri="sqlite:///redun.db")
    backend.load()
    assert backend.session

    min_compat_for_latest_version, _ = backend.get_db_version_required()

    # Ensure we are at current database version.
    version = backend.get_db_version()
    latest_version = backend.get_all_db_versions()[-1]
    assert version == latest_version

    prev_incompat_version = [
        v for v in backend.get_all_db_versions() if v < min_compat_for_latest_version
    ][-1]

    # Downgrade database to a version incompatible with the latest one.
    backend.migrate(prev_incompat_version)
    version = backend.get_db_version()
    assert version == prev_incompat_version

    # Load without migrating should leave version unchanged.
    with pytest.raises(RedunVersionError):
        backend.load(migrate=False)
    version = backend.get_db_version()
    assert version == prev_incompat_version

    # Load by default automigrates.
    backend.load()
    version = backend.get_db_version()
    assert version == latest_version


@use_tempdir
def test_db_automigrate_scheduler() -> None:
    """
    Scheduler should respect automigrate config variable.
    """
    scheduler = Scheduler(
        config=Config(
            {
                "backend": {
                    "db_uri": "sqlite:///redun.db",
                }
            }
        )
    )
    backend = cast(RedunBackendDb, scheduler.backend)
    assert backend.automigrate

    # Loading the database should automigrate by default.
    scheduler.load()

    # Ensure we are at current database version.
    version = backend.get_db_version()
    latest_version = backend.get_all_db_versions()[-1]
    assert version == latest_version

    # Downgrade database.
    prev_major_version = [v for v in backend.get_all_db_versions() if v < latest_version][-1]
    backend.migrate(prev_major_version)
    version = backend.get_db_version()
    assert version == prev_major_version

    # Get a new scheduler that does not do automigration.
    scheduler = Scheduler(
        config=Config({"backend": {"db_uri": "sqlite:///redun.db", "automigrate": "False"}})
    )
    backend = cast(RedunBackendDb, scheduler.backend)
    backend.create_engine()
    version = backend.get_db_version()
    assert version == prev_major_version


@use_tempdir
def test_automigrate_postgres() -> None:
    """
    Postgres databases should not automigrate.
    """
    File(".redun2/redun.ini").write(
        """
[backend]
db_uri = postgresql://host/db
"""
    )

    latest_version = REDUN_DB_VERSIONS[-1]

    scheduler = Scheduler(config=Config({"backend": {"db_uri": "postgresql://host/db"}}))
    with patch.object(RedunBackendDb, "migrate") as migrate_mock, patch.object(
        RedunBackendDb, "create_engine"
    ), patch.object(RedunBackendDb, "is_db_compatible", returns=True), patch.object(
        RedunBackendDb, "get_db_version", returns=latest_version
    ):
        scheduler.load()

    migrate_mock.assert_not_called()


def test_db_backend_clone() -> None:
    """
    Cloning should result in a new db backend that shares the same engine and task/execution
    records as the original, without sharing a database session
    """
    backend = RedunBackendDb(db_uri="sqlite:///:memory:")

    with pytest.raises(RedunDatabaseError, match="after.*load"):
        backend.clone()
    backend.load()
    clone = backend.clone()
    assert clone.engine is backend.engine
    assert clone.current_execution is backend.current_execution

    assert clone.session is not backend.session


def test_json_field(scheduler: Scheduler, session: Session) -> None:
    """
    Ensure we can properly store and filter a JSON db field.
    """

    # Add and filter a string value.
    session.add(
        Tag(
            tag_hash="111",
            entity_type=TagEntity.Value,
            entity_id="1",
            key="params",
            value="hello",
        )
    )
    session.commit()

    assert session.query(Tag).filter(Tag.value == sa_cast("hello", JSON)).one().value == "hello"
    assert not session.query(Tag).filter(Tag.value == sa_cast("bye", JSON)).all()

    # Add and filter an int value.
    session.add(
        Tag(tag_hash="222", entity_type=TagEntity.Value, entity_id="2", key="params", value=10)
    )
    session.commit()

    assert session.query(Tag).filter(Tag.value == sa_cast(10, JSON)).one().value == 10
    assert not session.query(Tag).filter(Tag.value == sa_cast(11, JSON)).all()

    # Add and filter a complex value.
    session.add(
        Tag(
            tag_hash="333",
            entity_type=TagEntity.Value,
            entity_id="2",
            key="params",
            value=[1, 2, 3],
        )
    )
    session.commit()

    assert session.query(Tag).filter(Tag.value == sa_cast([1, 2, 3], JSON)).one().value == [
        1,
        2,
        3,
    ]
    assert not session.query(Tag).filter(Tag.value == sa_cast([1, 2, 3, 4], JSON)).all()

    # Ensure dict normalization is used.
    session.add(
        Tag(
            tag_hash="444",
            entity_type=TagEntity.Value,
            entity_id="2",
            key="params",
            value={"a": 1, "b": 2},
        )
    )
    session.commit()

    assert session.query(Tag).filter(Tag.value == sa_cast({"b": 2, "a": 1}, JSON)).one().value == {
        "a": 1,
        "b": 2,
    }


class Data:
    def __init__(self, x: int):
        self.x = x


def test_value_no_class(scheduler: Scheduler, backend: RedunBackendDb, session: Session) -> None:
    """
    When loading a value without a importable class, use a preview.
    """

    # Store a value using a custom class, Data.
    value = Data(10)
    value_hash = backend.record_value(value)

    value_row = session.query(Value).filter(Value.value_hash == value_hash).one()

    # If data is not available, a PreviewValue should be returned.
    with patch.object(backend, "_get_value_data", lambda _: (None, False)):
        with pytest.warns(UserWarning, match="value_store_path"):
            value2 = value_row.preview
        assert isinstance(value2, PreviewValue)
        assert str(value2) == "redun.tests.test_db.Data(hash=2ad3f921, size=50)"

    # Remove custom class.
    original_Data = Data
    del globals()["Data"]

    # If class is not available, a PreviewClass should be returned.
    value3 = value_row.preview
    assert isinstance(value3, PreviewClass)
    assert str(value3) == "Data(x=10)"

    # Restore custom class.
    globals()["Data"] = original_Data



import dataclasses
import inspect
import io
import itertools
import json
import pickle
import re
import sys
import threading
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Iterable as IterableABC
from contextlib import contextmanager
from functools import lru_cache, wraps
from pickle import Unpickler
from pickle import dump as allowed_dump_func
from pickle import dumps as allowed_dumps_func
from typing import (
    IO,
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    NoReturn,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore


T = TypeVar("T")
NULL = object()
PICKLE_PROTOCOL = 3

# Additional python import paths added by user.
_redun_import_paths: List[str] = []

_local = threading.local()


def assert_never(value: NoReturn) -> NoReturn:
    """
    Helper method for exhaustive type checking.

    https://hakibenita.com/python-mypy-exhaustive-checking#type-narrowing-in-mypy
    """
    assert False, "Unhandled value: " + repr(value)


def str2bool(text: str) -> bool:
    """
    Parse a string into a bool.
    """
    text = text.lower()
    if text == "true":
        return True
    elif text == "false":
        return False
    else:
        raise ValueError(f"Cannot parse bool: '{text}'")


def add_import_path(path: str) -> None:
    """
    Add a path to the python import paths.
    """
    # Record redun import path.
    _redun_import_paths.insert(0, path)

    safe_path = path if path != "." else ""
    sys.path.insert(0, safe_path)


def get_import_paths() -> List[str]:
    """
    Returns extra import paths that have been added.
    """
    return _redun_import_paths


def clear_import_paths() -> None:
    """
    Clear all extra python import paths.
    """
    _redun_import_paths[:] = []


def json_dumps(value: Any) -> str:
    """
    Convert JSON-like values into a normalized string.

    Keys are sorted and no whitespace is used around delimiters.
    """
    return json.dumps(value, separators=(",", ":"), sort_keys=True)


def pickle_dump(obj: Any, file: IO) -> None:
    """
    Official pickling method for redun.

    This is used to standardize the protocol used for serialization.
    """
    return allowed_dump_func(obj, file, protocol=PICKLE_PROTOCOL)


def pickle_dumps(obj: Any) -> bytes:
    """
    Official pickling method for redun.

    This is used to standardize the protocol used for serialization.
    """
    return allowed_dumps_func(obj, protocol=PICKLE_PROTOCOL)


def pickle_loads(data: bytes) -> Any:
    """
    Official unpickling method for redun.

    This is used to allow pickle previews with a preview context is active.
    """
    if getattr(_local, "num_pickle_preview_active", 0) > 0:
        return pickle_preview(data, raise_error=_local.pickle_preview_raise_error)
    else:
        return pickle.loads(data)


@contextmanager
def with_pickle_preview(raise_error: bool = False) -> Iterator[None]:
    """
    Enable pickle preview within a context.

    We keep a count of how many nested contexts are active.
    """
    num_pickle_preview_active = getattr(_local, "num_pickle_preview_active", 0)
    _local.num_pickle_preview_active = num_pickle_preview_active + 1
    _local.pickle_preview_raise_error = raise_error
    yield
    _local.num_pickle_preview_active -= 1


def iter_nested_value_children(value: Any) -> Iterable[Tuple[bool, Any]]:
    """
    Helper function that iterates through the children of a possibly nested value.

    Yields: (is_leaf: Bool, value: Any)
    """

    value_type = type(value)

    # Note: this works for list, set, dict, tuple, namedtuple.
    if value_type in (list, tuple, set) or isinstance(value, tuple) and hasattr(value, "_fields"):
        for item in value:
            yield False, item

    elif value_type == dict:
        for key in value.keys():
            yield False, key
        for val in value.values():
            yield False, val

    elif dataclasses.is_dataclass(value_type):
        for field in dataclasses.fields(value):
            yield False, value.__dict__[field.name]

    else:
        # Visit leaf values.
        yield True, value


def iter_nested_value(value: Any) -> Iterable[Any]:
    """
    Iterate through the leaf values of a nested value.
    """
    stack = [(False, value)]
    while stack:
        is_leaf, value = stack.pop()
        if is_leaf:
            yield value
        else:
            stack.extend(iter_nested_value_children(value))


def map_nested_value(func: Callable, value: Any) -> Any:
    """
    Map a func to the leaves of a nested value.
    """
    value_type = type(value)

    if value_type == list:
        return [map_nested_value(func, item) for item in value]

    elif value_type == tuple:
        return tuple([map_nested_value(func, item) for item in value])

    elif isinstance(value, tuple) and hasattr(value, "_fields"):
        # Namedtuple.
        return value_type(*[map_nested_value(func, item) for item in value])

    elif value_type == set:
        return {map_nested_value(func, item) for item in value}

    elif value_type == dict:
        return {
            map_nested_value(func, key): map_nested_value(func, val) for key, val in value.items()
        }

    elif dataclasses.is_dataclass(value_type):
        mapped_value = value_type(
            **{
                field.name: map_nested_value(func, value.__dict__[field.name])
                for field in dataclasses.fields(value)
            }
        )
        # Copy over any non-field items from the origin value __dict__  (such as __orig_class__,
        # which exists for subscripted generic objects) that haven't made it to the mapped value.
        mapped_value.__dict__ = {
            **value.__dict__,
            **mapped_value.__dict__,
        }
        return mapped_value

    else:
        return func(value)


def trim_string(text: str, max_length: int = 200, ellipsis: str = "...") -> str:
    """
    If text is longer than max_length then return a trimmed string.
    """
    assert max_length >= len(ellipsis)
    if len(text) > max_length:
        return text[: max_length - len(ellipsis)] + ellipsis
    else:
        return text


def merge_dicts(dicts: List[T]) -> T:
    """
    Merge a list of (nested) dicts into a single dict.
    .. code-block:: python
        assert merge_dicts([
            {"a": 1, "b": 2},
            {"b": 3, "c": 4},
        ]) == {"a": 1, "b": 3, "c": 4}
        assert merge_dicts([
            {"a": {"a1": 1}},
            {"a": {"a2": 2}}
        ]) == {"a": {"a1": 1, "a2": 2}}
    """
    if len(dicts) == 1:
        return dicts[0]

    elif any(not isinstance(dct, dict) for dct in dicts):
        # For non-dicts, last value takes precedence.
        return dicts[-1]

    else:
        # Group by keys.
        key2values = defaultdict(list)
        for dct in dicts:
            # At this point, we can safely assume dct has type dict, so use cast.
            for key, value in cast(dict, dct).items():
                key2values[key].append(value)

        # Recursively merge values.
        return cast(T, {key: merge_dicts(values) for key, values in key2values.items()})


def json_cache_key(args: tuple, kwargs: dict) -> str:
    """
    Returns a cache key for arguments that are JSON compatible.
    """
    return json.dumps([args, kwargs], sort_keys=True)


class CacheArgs:
    """
    Arguments for a cached function.
    """

    def __init__(self, cache_key, args, kwargs):
        self.cache_key = cache_key
        self.args = args
        self.kwargs = kwargs

    def __hash__(self):
        """
        The normal @lru_cache will call this to build a cache key, so we
        can inject custom cache key (e.g. json_cache_key) here.
        """
        return hash(self.cache_key(self.args, self.kwargs))


def lru_cache_custom(maxsize: int, cache_key=lambda x: x):
    """
    LRU cache with custom cache key.
    """

    @lru_cache(maxsize=maxsize)
    def call_func(func: Callable[..., T], args: CacheArgs) -> T:
        return func(*args.args, **args.kwargs)

    def deco(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            return call_func(func, CacheArgs(cache_key, args, kwargs))

        return wrapped

    return deco


def get_func_source(func: Callable) -> str:
    """
    Return the source code for a function.
    """
    source = inspect.getsource(func)

    # Try to trim away the decorators.
    lines = source.split("\n")
    for i, line in enumerate(lines):
        if re.match(r"^ *def ", line):
            return "\n".join(lines[i:])

    return source


def format_table(table: List[List], justs: str, min_width: int = 0) -> Iterator[str]:
    """
    Format table with justified columns.
    """
    assert len(justs) == len(table[0])

    # Convert to strings.
    table = [[str(cell) for cell in row] for row in table]

    column_widths = [
        max(max(len(table[i][j]) for i in range(len(table))), min_width)
        for j in range(len(table[0]))
    ]

    def justify(text, just, width):
        if just == "l":
            return text.ljust(width)
        elif just == "r":
            return text.rjust(width)
        raise NotImplementedError(just)

    for i, row in enumerate(table):
        if i == 1:
            yield ""
        yield " ".join(
            justify(cell, just, width) for cell, just, width in zip(row, justs, column_widths)
        )


class PreviewClass(ABC):
    """
    Generic class to use for unknown classes in a pickle
    """

    _module = "module"
    _name = "name"

    def __init__(self, *args, **kwargs):
        self._preview_args = args
        self.__dict__.update(kwargs)

    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls)
        self._preview_args = args
        self.__dict__.update(kwargs)
        return self

    def __repr__(self):
        attr = []
        if hasattr(self, "_preview_args"):
            # Object was instantiated with constructor or with __new__.
            attr.extend(map(repr, self._preview_args))

        # Object was populated through __dict__.
        attr.extend(
            f"{key}={repr(value)}"
            for key, value in self.__dict__.items()
            if key != "_preview_args"
        )
        return f"{self._name}({', '.join(attr)})"

    def __getitem__(self, index: int) -> Any:
        """
        Mimic tuple type.
        """
        return self._preview_args[index]

    # This is explicitly defined so mypy type checking works on PreviewClass instances.
    def __getattr__(self, name: str) -> Any:
        if name in self.__dict__:
            return self.__dict__[name]
        raise AttributeError(f"{name} not an attribute of {self._name} PreviewClass")


class PreviewingUnpickler(Unpickler):
    """
    A specialized unpickler that allows previewing of class attributes for
    unknown classes.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Registry of stub classes for unknown classes.
        self._classes: Dict[str, Type] = {}

    def find_class(self, module: str, name: str) -> Type:
        try:
            return super().find_class(module, name)

        except (AttributeError, ModuleNotFoundError):
            class_name = f"{module}.{name}"
            cached_class = self._classes.get(class_name)
            if cached_class:
                return cached_class

            class PreviewSubclass(PreviewClass):
                _module = module
                _name = name

            PreviewSubclass.__name__ = name
            PreviewSubclass.__module__ = module

            self._classes[class_name] = PreviewSubclass
            return PreviewSubclass


class PreviewUnpicklable:
    def __init__(self, error):
        self.error = error

    def __repr__(self) -> str:
        return "Unpicklable('{}')".format(self.error)


def pickle_preview(data: bytes, raise_error: bool = False) -> Any:
    """
    Loads a pickle file, falling back to previewing objects as needed.

    For example, if the class is unknown, will return a PreviewClass
    subclass that can be used to peek at attributes.
    """
    infile = io.BytesIO(data)
    unpickler = PreviewingUnpickler(infile)

    try:
        with with_pickle_preview(raise_error):
            return unpickler.load()
    except Exception as error:
        if raise_error:
            raise
        else:
            return PreviewUnpicklable(error)


class Comparable(Protocol):
    """Protocol for annotating comparable types."""

    @abstractmethod
    def __lt__(self: T, other: T) -> bool:
        pass


Key = TypeVar("Key", bound="Comparable")
Value = TypeVar("Value", bound="Comparable")


def _multidict2items(dct: Dict[Key, List[Value]]) -> List[Tuple[Key, Value]]:
    return [(key, value) for key, values in dct.items() for value in values]


class MultiMap(Generic[Key, Value]):
    """
    An associative array with repeated keys.
    """

    def __init__(self, items: Iterable[Union[Tuple[Key, Value], List]] = ()):
        self._data: Dict[Key, List[Value]] = {}
        self._len = 0
        for key, value in items:
            if key not in self._data:
                self._data[key] = []
            self._data[key].append(value)
            self._len += 1

    def __repr__(self) -> str:
        return "MultiMap({})".format(list(self.items()))

    def __len__(self) -> int:
        return self._len

    def __getitem__(self, key: Key) -> List[Value]:
        return self._data[key]

    def __iter__(self) -> Iterator[Tuple[Key, Value]]:
        for key, values in self._data.items():
            for value in values:
                yield (key, value)

    def __contains__(self, key: Key) -> bool:
        return key in self._data

    def _equal_dicts(self, dict1: Dict[Key, List[Value]], dict2: Dict[Key, List[Value]]) -> bool:
        return sorted(_multidict2items(dict1)) == sorted(_multidict2items(dict2))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, MultiMap):
            return self._equal_dicts(self._data, other._data)

        elif isinstance(other, dict):
            return self._equal_dicts(self._data, other)

        elif isinstance(other, IterableABC):
            return sorted(self) == sorted(other)

        else:
            return False

    def has_item(self, key: Key, value: Value) -> bool:
        return value in self._data.get(key, ())

    def get(self, key: Key, default: Any = NULL) -> List[Value]:
        if default is NULL:
            default = []
        return self._data.get(key, default)

    def keys(self) -> Iterator[Key]:
        return iter(self._data.keys())

    def values(self) -> Iterator[Value]:
        return itertools.chain.from_iterable(self._data.values())

    def items(self) -> Iterator[Tuple[Key, Value]]:
        return iter(self)

    def as_dict(self) -> Dict[Key, List[Value]]:
        return self._data

    def add(self, key: Key, value: Value) -> None:
        if key not in self._data:
            self._data[key] = []
        self._data[key].append(value)
        self._len += 1



import os

from setuptools import find_packages, setup

REQUIRE_POSTGRES = os.getenv("REQUIRE_POSTGRES") == "1"
PSYCOPG2_VERSION = "psycopg2>=2.8"

requirements = [
    # By using the extra deps boto3 and awscli, we help the solver find
    # a solution faster, since aiobotocore currently requires pinned dependencies of
    # boto3 and awscli.
    "aiobotocore[boto3,awscli]>=2.0.1",
    "aiohttp>=3.7.4,<4",
    "alembic>=1.4",
    "boto3>=1.16.63",
    # Temporarily cap botocore version until bug is fixed
    # https://github.com/iterative/dvc/issues/8513#issuecomment-1298761683
    "botocore>=1.22.8,<1.28.0",
    "gcsfs>=2021.4.0",
    "s3fs>=2021.11.1",
    # Using 2.1 instead of 3.0 in case future 2.x versions drop support
    # for some legacy APIs still supported in 2.0
    "sqlalchemy>=1.4.0,<2.1",
    "python-dateutil>=2.8",
    # If updating this list, check executors/aws_glue.py stays up to date with
    # packages needed to run in the glue environment.
    # 3.6 backports
    'dataclasses>=0.8;python_version<="3.6"',
    'types-dataclasses>=0.6.6;python_version<="3.6"',
]

extras = {
    "glue": ["pandas", "pyarrow", "pyspark"],
    "k8s": "kubernetes>=22.6",
    "viz": "pygraphviz",
}

if REQUIRE_POSTGRES:
    requirements.append(PSYCOPG2_VERSION)
else:
    extras["postgres"] = [PSYCOPG2_VERSION]


def get_version() -> str:
    """
    Get the redun package version.
    """
    # Technique from: https://packaging.python.org/guides/single-sourcing-package-version/
    basedir = os.path.dirname(__file__)
    module_path = os.path.join(basedir, "redun", "__init__.py")
    with open(module_path) as infile:
        for line in infile:
            if line.startswith("__version__"):
                _, version, _ = line.split('"', 2)
                return version
    assert False, "Cannot find redun package version"


setup(
    name="redun",
    version=get_version(),
    zip_safe=True,
    packages=find_packages(),
    author="Matt Rasmussen",
    author_email="rasmus@insitro.com",
    url="https://github.com/insitro/redun/",
    description="Yet another redundant workflow engine.",
    long_description="""
redun aims to be a more expressive and efficient workflow framework, built on
top of the popular Python programming language. It takes the somewhat contrarian
view that writing dataflows directly is unnecessarily restrictive, and by doing
so we lose abstractions we have come to rely on in most modern high-level
languages (control flow, compositiblity, recursion, high order functions, etc).
redun's key insight is that workflows can be expressed as lazy expressions, that
are then evaluated by a scheduler which performs automatic parallelization,
caching, and data provenance logging.

redun's key features are:

- Workflows are defined by lazy expressions that when evaluated emit dynamic directed acyclic
  graphs (DAGs), enabling complex data flows.
- Incremental computation that is reactive to both data changes as well as code changes.
- Workflow tasks can be executed on a variety of compute backend (threads, processes, AWS batch
  jobs, Spark jobs, etc).
- Data changes are detected for in memory values as well as external data sources such as files
  and object stores using file hashing.
- Code changes are detected by hashing individual Python functions and comparing against
  historical call graph recordings.
- Past intermediate results are cached centrally and reused across workflows.
- Past call graphs can be used as a data lineage record and can be queried for debugging and
  auditing.
    """,
    scripts=["bin/redun"],
    include_package_data=True,
    python_requires=">= 3.6",
    install_requires=requirements,
    extras_require=extras,
)



from functools import reduce
from itertools import islice
from typing import Any, Callable, Dict, Iterable, Iterator, List, Optional, Set, Tuple

import sqlalchemy as sa
from sqlalchemy.orm.query import Query
from sqlalchemy.sql.expression import cast as sa_cast
from sqlalchemy.sql.expression import select

from redun.backends.db import (
    JSON,
    Argument,
    Base,
    CallNode,
    Execution,
    File,
    Job,
    Subvalue,
    Tag,
    Task,
    Value,
)
from redun.tags import ANY_VALUE


class CallGraphQuery:
    """
    Query class for efficiently querying across all models in a CallGraph.
    """

    # Look up for model primary keys.
    MODEL_PKS = {
        Execution: "id",
        Job: "id",
        CallNode: "call_hash",
        Value: "value_hash",
        Task: "hash",
    }
    MODEL_CLASSES = {
        "Execution": Execution,
        "Job": Job,
        "CallNode": CallNode,
        "Task": Task,
        "Value": Value,
    }
    MODEL_NAMES = list(MODEL_CLASSES.keys())

    ExecTag = sa.orm.aliased(Tag)

    def __init__(
        self,
        session: sa.orm.Session,
        joins: Optional[Set[str]] = None,
        execution_joins: Optional[List[Callable[[Query], Query]]] = None,
        filters: Optional[List] = None,
        order_by: Optional[str] = None,
        filter_types: Optional[Set] = None,
        executions: Optional[Query] = None,
        jobs: Optional[Query] = None,
        call_nodes: Optional[Query] = None,
        tasks: Optional[Query] = None,
        values: Optional[Query] = None,
        value_subqueries: Optional[List[Query]] = None,
    ):
        self._session = session
        self._joins: Set[str] = joins or set()
        self._execution_joins: List[Callable[[Query], Query]] = execution_joins or []
        self._filters: List = filters or []
        self._order_by = order_by
        self._filter_types = filter_types if filter_types is not None else set(self.MODEL_NAMES)

        # Subqueries.
        self._executions = executions or self._session.query(Execution)
        self._jobs = jobs or self._session.query(Job)
        self._call_nodes = call_nodes or self._session.query(CallNode)
        self._tasks = tasks or self._session.query(Task)
        self._values = values if values is not None else self._session.query(Value)
        self._value_subqueries = value_subqueries

    @property
    def subqueries(self) -> Iterable[Tuple[str, Query]]:
        """
        Iterates through all subqueries.

        Yields
        ------
        (type_name, subquery) : Tuple[str, Query]
        """
        return [
            ("Execution", self._executions),
            ("Job", self._jobs),
            ("CallNode", self._call_nodes),
            ("Task", self._tasks),
            ("Value", self._values),
        ]

    def clone(self, **kwargs: Any) -> "CallGraphQuery":
        """
        Returns a clone of the query with updates specified by `kwargs`.
        """
        clone_kwargs: Dict[str, Any] = {
            "joins": self._joins,
            "execution_joins": self._execution_joins,
            "filters": self._filters,
            "order_by": self._order_by,
            "filter_types": self._filter_types,
            "executions": self._executions,
            "jobs": self._jobs,
            "call_nodes": self._call_nodes,
            "tasks": self._tasks,
            "values": self._values,
            "value_subqueries": self._value_subqueries,
        }
        clone_kwargs.update(kwargs)
        return CallGraphQuery(self._session, **clone_kwargs)

    def filter_types(self, types: Iterable[str]) -> "CallGraphQuery":
        """
        Filter query by record type.
        """
        return self.clone(filter_types=set(types))

    def filter_ids(self, _ids: Iterable[str]) -> "CallGraphQuery":
        """
        Filter query by record ids.
        """
        ids = list(_ids)

        def filter(query):
            return query.clone(
                executions=query._executions.filter(Execution.id.in_(ids)),
                jobs=query._jobs.filter(Job.id.in_(ids)),
                call_nodes=query._call_nodes.filter(CallNode.call_hash.in_(ids)),
                tasks=query._tasks.filter(Task.hash.in_(ids)),
                values=query._values.filter(Value.value_hash.in_(ids)),
            )

        return self.clone(filters=self._filters + [filter])

    def like_id(self, id: str) -> "CallGraphQuery":
        """
        Filter query by record id prefix `id`.
        """

        def filter(query):
            pattern = id + "%"
            return query.clone(
                executions=query._executions.filter(Execution.id.like(pattern)),
                jobs=query._jobs.filter(Job.id.like(pattern)),
                call_nodes=query._call_nodes.filter(CallNode.call_hash.like(pattern)),
                tasks=(
                    self._tasks.filter(Task.hash.like(pattern)).union(
                        self._tasks.filter(Task.name == id)
                    )
                ),
                values=query._values.filter(Value.value_hash.like(pattern)),
            )

        return self.clone(filters=self._filters + [filter])

    def _join_files(self):
        return self.clone(values=self._values.join(File, File.value_hash == Value.value_hash))

    def _join_jobs(self):
        return self.clone(executions=self._executions.join(Job, Job.id == Execution.job_id))

    def _join_tasks(self):
        return self.clone(
            jobs=self._jobs.join(Task, Task.hash == Job.task_hash),
        )

    def _join_executions(self):
        """
        Join Executions to each subquery.
        """
        query = self

        # Values can be joined to Execution via four possible paths. Thus far,
        # we have found it most efficient to perform all four paths and union
        # their results.

        # Values can connect to CallNodes either directly as a result or
        # through Argument as an argument. Values can also be a subvalue of a
        # parent Value. These possibilities combine to give four join paths.
        value_results = query._values.join(CallNode, CallNode.value_hash == Value.value_hash).join(
            Job, Job.call_hash == CallNode.call_hash
        )
        value_args = (
            query._values.join(Argument, Argument.value_hash == Value.value_hash)
            .join(CallNode, CallNode.call_hash == Argument.call_hash)
            .join(Job, Job.call_hash == CallNode.call_hash)
        )
        subvalue_results = (
            query._values.join(Subvalue, Subvalue.value_hash == Value.value_hash)
            .join(CallNode, CallNode.value_hash == Subvalue.parent_value_hash)
            .join(Job, Job.call_hash == CallNode.call_hash)
        )
        subvalue_args = (
            query._values.join(Subvalue, Subvalue.value_hash == Value.value_hash)
            .join(Argument, Argument.value_hash == Subvalue.parent_value_hash)
            .join(CallNode, CallNode.call_hash == Argument.call_hash)
            .join(Job, Job.call_hash == CallNode.call_hash)
        )
        subqueries = [value_results, value_args, subvalue_results, subvalue_args]

        # Perform additional joins on execution.
        for i, _ in enumerate(subqueries):
            for join in self._execution_joins:
                subqueries[i] = join(subqueries[i])

        return query.clone(
            executions=query._executions,
            jobs=query._jobs,
            call_nodes=(query._call_nodes.join(Job, Job.call_hash == CallNode.call_hash)),
            tasks=(query._tasks.join(Job, Job.task_hash == Task.hash)),
            value_subqueries=subqueries,
        )

    def filter_execution_ids(self, execution_ids: Iterable[str]) -> "CallGraphQuery":
        """
        Filter query by Execution ids.
        """
        execution_ids = list(execution_ids)

        def filter(query):
            return query.clone(
                executions=query._executions.filter(Execution.id.in_(execution_ids)),
                jobs=query._jobs.filter(Job.execution_id.in_(execution_ids)),
                call_nodes=query._call_nodes.filter(Job.execution_id.in_(execution_ids)),
                tasks=query._tasks.filter(Job.execution_id.in_(execution_ids)),
                values=query._values.filter(Job.execution_id.in_(execution_ids)),
            )

        return self.clone(
            joins=self._joins | {"execution"},
            filters=self._filters + [filter],
        )

    def filter_execution_statuses(self, execution_statuses: Iterable[str]) -> "CallGraphQuery":
        """
        Filter by Execution status.
        """
        assert execution_statuses

        def term(status):
            if status == "DONE":
                return Job.end_time.isnot(None)
            elif status == "FAILED":
                return Job.end_time.is_(None)
            else:
                raise NotImplementedError(status)

        execution_clause = reduce(sa.or_, map(term, execution_statuses))

        def filter(query):
            return query.clone(executions=query._executions.filter(execution_clause))

        return self.clone(
            filter_types=self._filter_types & {"Execution"},
            joins=self._joins | {"job"},
            filters=self._filters + [filter],
            order_by="time",
        )

    def filter_job_statuses(self, job_statuses: Iterable[str]) -> "CallGraphQuery":
        """
        Filter by Job status.
        """
        assert job_statuses

        def job_term(status):
            if status == "FAILED":
                return Job.end_time.is_(None)
            elif status == "DONE":
                return Job.end_time.isnot(None) & Job.cached.is_(False)
            elif status == "CACHED":
                return Job.cached.is_(True)
            else:
                raise NotImplementedError(status)

        job_clause = reduce(sa.or_, map(job_term, job_statuses))

        def filter(query):
            return query.clone(jobs=query._jobs.filter(job_clause))

        return self.clone(
            filter_types=self._filter_types & {"Job"},
            filters=self._filters + [filter],
        )

    def filter_task_names(self, task_names: Iterable[str]) -> "CallGraphQuery":
        """
        Filter by Task names.
        """

        def task_term(task_name: str):
            if "." in task_name:
                namespace, name = task_name.rsplit(".", 1)
                return (Task.namespace == namespace) & (Task.name == name)
            else:
                return Task.name == task_name

        clause = reduce(sa.or_, map(task_term, task_names))

        def filter(query):
            return query.clone(
                jobs=query._jobs.filter(clause),
                tasks=query._tasks.filter(clause),
            )

        return self.clone(
            filter_types=self._filter_types & {"Job", "Task"},
            joins=self._joins | {"task"},
            filters=self._filters + [filter],
        )

    def filter_value_types(self, value_types: Iterable[str]) -> "CallGraphQuery":
        """
        Filter query by Value types.
        """

        def filter(query):
            return query.clone(values=query._values.filter(Value.type.in_(value_types)))

        return self.clone(
            filter_types=self._filter_types & {"Value"},
            filters=self._filters + [filter],
        )

    def filter_file_paths(self, paths: Iterable[str]) -> "CallGraphQuery":
        """
        Filter by File.path patterns.

        `paths` can contain "*" to perform wildcard matching.
        """
        # Convert path patterns into db LIKE patterns.
        like_patterns = [path.replace("*", "%") for path in paths]

        def filter(query):
            return query.clone(
                values=query._values.filter(
                    sa.or_(File.path.like(like_pattern) for like_pattern in like_patterns)
                )
            )

        return self.clone(
            filter_types=self._filter_types & {"Value"},
            joins=self._joins | {"file"},
            filters=self._filters + [filter],
        )

    def _query_filter_tags(
        self, query: Query, entity_id_col: Any, table: Any, tags: Iterable[Tuple[str, Any]]
    ) -> Query:
        """
        Build query for filtering tags.
        """
        for key, value in tags:
            tag_query = self._session.query(table).filter(
                table.is_current.is_(True), table.key == key
            )
            if value is not ANY_VALUE:
                tag_query = tag_query.filter(table.value == sa_cast(value, JSON))
            # We perform this filter as a subquery in order to find entities
            # with multiple matching tags.
            tag_query = tag_query.subquery()
            query = query.join(tag_query, tag_query.c.entity_id == entity_id_col)
        return query

    def filter_tags(self, tags: Iterable[Tuple[str, Any]]) -> "CallGraphQuery":
        """
        Filter by tags.
        """

        def filter(query):
            return query.clone(
                executions=self._query_filter_tags(query._executions, Execution.id, Tag, tags),
                jobs=self._query_filter_tags(query._jobs, Job.id, Tag, tags),
                call_nodes=self._query_filter_tags(
                    query._call_nodes, CallNode.call_hash, Tag, tags
                ),
                tasks=self._query_filter_tags(query._tasks, Task.hash, Tag, tags),
                values=self._query_filter_tags(query._values, Value.value_hash, Tag, tags),
            )

        return self.clone(
            filters=self._filters + [filter],
        )

    def filter_execution_tags(self, tags: Iterable[Tuple[str, Any]]) -> "CallGraphQuery":
        """
        Filter by tag on executions.
        """

        def filter(query):
            # Values will be filtered by the execution_join below.
            return query.clone(
                executions=self._query_filter_tags(
                    query._executions, Execution.id, self.ExecTag, tags
                ),
                jobs=self._query_filter_tags(query._jobs, Job.execution_id, self.ExecTag, tags),
                call_nodes=self._query_filter_tags(
                    query._call_nodes, Job.execution_id, self.ExecTag, tags
                ),
                tasks=self._query_filter_tags(query._tasks, Job.execution_id, self.ExecTag, tags),
            )

        def exec_join(sa_query: Query) -> Query:
            return self._query_filter_tags(sa_query, Job.execution_id, self.ExecTag, tags)

        return self.clone(
            joins=self._joins | {"execution"},
            execution_joins=self._execution_joins + [exec_join],
            filters=self._filters + [filter],
        )

    def order_by(self, order_by: str) -> "CallGraphQuery":
        """
        Order query.

        order_by: str
            The only supported value is 'time'.
        """
        return self.clone(order_by=order_by)

    def build(self):
        """
        Apply all joins and filters to subqueries.
        """
        query = self

        # Perform joins.
        if "file" in self._joins:
            query = query._join_files()
        if "job" in self._joins:
            query = query._join_jobs()
        if "task" in self._joins:
            query = query._join_tasks()
        if "execution" in self._joins:
            query = query._join_executions()

        # Perform filters.
        for filter in self._filters:
            query = filter(query)

        # Filter and union value subqueries.
        if query._value_subqueries:
            subqueries = [self.clone(values=subquery) for subquery in query._value_subqueries]
            for i, _ in enumerate(subqueries):
                for filter in self._filters:
                    subqueries[i] = filter(subqueries[i])
            values = subqueries[0]._values.union(*(q._values for q in subqueries[1:]))
            query = query.clone(values=values)

        # Perform order_by.
        if self._order_by == "time":
            Job2 = sa.orm.aliased(Job)
            query = query.clone(
                executions=(
                    query._executions.add_columns(Job2.start_time)
                    .join(Job2, Execution.job_id == Job2.id)
                    .order_by(Job2.start_time.desc())
                ),
                jobs=(query._jobs.order_by(Job.start_time.desc())),
            )
        elif self._order_by:
            raise NotImplementedError(self._order_by)

        # Remove filters, joins, etc since they are now built in.
        return query.clone(filters=[], joins={}, order_by=None)

    def empty(self) -> "CallGraphQuery":
        """
        Returns an empty query.
        """
        query = CallGraphQuery(self._session)
        return self.clone(
            executions=query._executions.filter(False),
            jobs=query._jobs.filter(False),
            call_nodes=query._call_nodes.filter(False),
            tasks=query._tasks.filter(False),
            values=query._values.filter(False),
        )

    def all(self) -> Iterator[Base]:
        """
        Yields all records matching query.
        """
        query = self.build()

        for record_type, subquery in query.subqueries:
            if record_type in self._filter_types:
                # Normally, `subquery.statement.distinct()` would be enough here
                # but the extra select() wrapper is needed when using unions.
                # Also, select() doesn't accept string names for now, so we need to look up
                # the actual type.
                selectable = select(self.MODEL_CLASSES[record_type]).from_statement(
                    subquery.statement.distinct()
                )
                yield from self._session.execute(selectable).scalars()

    def one(self) -> Base:
        """
        Returns exactly one record. Raises error if too few or too many.
        """
        [result] = self.all()
        return result

    def one_or_none(self) -> Base:
        """
        Returns exactly one record. Returns None if too few or too many.
        """
        results = list(self.all())
        if not results:
            return None
        [result] = results
        return result

    def first(self) -> Optional[Base]:
        """
        Returns first record if it exists.
        """
        return next(self.all(), None)

    def limit(self, size) -> Iterator[Base]:
        """
        Yields at most `size` records from query.
        """
        query = self.build().clone(
            executions=self._executions.limit(size),
            jobs=self._jobs.limit(size),
            call_nodes=self._call_nodes.limit(size),
            tasks=self._tasks.limit(size),
            values=self._values.limit(size),
        )
        yield from islice(query.all(), 0, size)

    def count(self) -> Iterator[Tuple[str, int]]:
        """
        Returns counts for each record type.
        """
        query = self.build()
        for record_type, subquery in query.subqueries:
            if record_type in self._filter_types:
                yield record_type, subquery.distinct().count()

    def select(self, *columns: Iterable[str], flat: bool = False) -> Iterator[Any]:
        """
        Select columns to return in query.
        """
        for record in self.all():
            row = []
            for column in columns:
                if column == "id":
                    pk = self.MODEL_PKS[type(record)]
                    row.append(getattr(record, pk))

                elif column == "type":
                    row.append(type(record).__name__)

                else:
                    raise ValueError(f"Unknown column '{column}'.")

            if flat:
                [value] = row
                yield value
            else:
                yield tuple(row)



import os
from configparser import ConfigParser, ExtendedInterpolation, SectionProxy
from typing import Any, Dict, Iterable, Optional, Union

from redun.file import File


class RedunExtendedInterpolation(ExtendedInterpolation):
    """
    When performing variable interpolation fallback to environment variables.

    For example, if the environment variable ROLE is defined, we can reference
    it in the `redun.ini` file as follows:

    .. code-block:: ini
        [executors.batch]
        role = ${ROLE}
    """

    def before_get(self, parser, section, option, value, defaults):
        # Fallback to environment variables when interpolating variables.
        defaults = {
            **defaults,
            **os.environ,
        }
        return super().before_get(parser, section, option, value, defaults)


class RedunConfigParser(ConfigParser):
    def optionxform(self, optionstr):
        # Treat option names as case sensitive.
        return optionstr


class Config:
    """
    Extends ConfigParser to support nested sections.
    """

    def __init__(self, config_dict: Optional[dict] = None):
        self.parser = RedunConfigParser(interpolation=RedunExtendedInterpolation())
        self._sections: "Section" = {}
        if config_dict:
            self.read_dict(config_dict)

    def read_string(self, string: str) -> None:
        self.parser.read_string(string)
        self._sections = self._parse_sections(self.parser)

    def read_path(self, filename: str) -> None:
        with File(filename).open() as infile:
            self.parser.read_file(infile)
            self._sections = self._parse_sections(self.parser)

    def read_dict(self, config_dict: dict) -> None:
        self.parser.read_dict(config_dict)
        self._sections = self._parse_sections(self.parser)

    def _parse_sections(self, parser) -> Union[dict, "Section"]:
        """
        Parse a dot notation section into nested dicts.
        """
        full_sections = parser.sections()
        nested_sections: dict = {}
        for full_section in full_sections:
            parts = full_section.split(".")
            ptr = nested_sections
            for part in parts[:-1]:
                if part not in ptr:
                    ptr[part] = {}
                ptr = ptr[part]

            ptr[parts[-1]] = parser[full_section]
        return nested_sections

    def get(self, key: str, default: Any = None) -> Any:
        return self._sections.get(key, default)

    def __getitem__(self, section_name: str) -> Any:
        return self._sections[section_name]

    def __setitem__(self, section_name: str, section: "Section") -> "Section":
        # TODO: See how to properly get type checking for this assignment.
        self._sections[section_name] = section  # type: ignore
        return section

    def keys(self) -> Iterable[str]:
        return self._sections.keys()

    def items(self):
        return self._sections.items()

    def get_config_dict(self, replace_config_dir=None) -> Dict[str, Dict]:
        """Return a python dict that can be used to reconstruct the Config object

        config2 = Config(config_dict=config1.get_config_dict()) should result in identical
        Configs config1 and config2.

        Note: The structure of the returned Dict is essentially a two-level dict corresponding to
        INI file structure. Top-level key is a dot-separated section name.  Top-level value is a
        single-level dict containing key/values for a single section.

        Parameters
        ----------
        replace_config_dir
            if not None, replaces any variables containing the machine-local config_dir (as
            obtained by `redun.cli.get_config_dir()` with `replace_config_dir`.
            Typical values that are replaced are `backend::config_dir`, `db_uri`,
            `repos.default::config_dir`.

        Returns
        -------
        A copy of this object's config in two-level format, optionally with values containing
        the local config directory replaced.
        """
        from redun.cli import get_config_dir

        result = {}
        local_config_dir = get_config_dir()

        def substitute_config_dir(s):
            # Replace string values containing local config_dir with the provided one
            if replace_config_dir is not None and isinstance(s, str):
                return s.replace(local_config_dir, replace_config_dir)
            return s

        def convert_to_dict(path, obj):
            if isinstance(obj, SectionProxy):
                result[path] = {k: substitute_config_dir(v) for k, v in obj.items()}
                return
            for key in obj.keys():
                convert_to_dict(f"{path}.{key}" if path else key, obj[key])

        convert_to_dict("", self)
        return result


Section = Union[dict, SectionProxy, Config]


def create_config_section(config_dict: Optional[dict] = None) -> SectionProxy:
    """
    Create a default section.
    """
    return Config({"section": config_dict or {}})["section"]



import os.path
import sys

import pygraphviz as pgv

import redun.visualize
from redun import File
from redun.backends.db import CallNode, Execution, Job, RedunBackendDb, Value
from redun.backends.db.query import CallGraphQuery
from redun.cli import RedunClient
from redun.visualize import MAX_LABEL_LEN, SEPARATOR

DEFAULT_WORKFLOW = """
from redun import task

@task()
def task1(x: str, y: str):
    return x + y

@task()
def main(x: str = "hi", y: str = "bye"):
    return task1(x, y)
"""


def run_workflow(argv, file_contents=DEFAULT_WORKFLOW):
    sys.modules.pop("workflow", None)
    file = File("workflow.py")
    file.write(file_contents)

    client = RedunClient()
    client.execute(argv)

    parser = client.get_command_parser()
    args, _ = parser.parse_known_args(argv[1:])

    backend = client.get_backend(args)
    assert isinstance(backend, RedunBackendDb)
    exec_id = (
        backend.session.query(Execution)
        .join(Job, Job.id == Execution.job_id)
        .order_by(Job.start_time.desc())
        .first()
        .id
    )
    return client, exec_id, file


def test_basic_viz():
    file_contents = """
from redun import task

@task()
def main(x: str = "hi", y: bool = True):
    return (x, y)
"""

    long_hi = "hi" * 100
    argv = ["redun", "run", "workflow.py", "main", "--x", long_hi]

    client, exec_id, file = run_workflow(argv, file_contents)

    viz_argv = [
        "redun",
        "viz",
        exec_id,
    ]
    client.execute(viz_argv)

    dot_file, image = File(f"{exec_id}.dot"), File(f"{exec_id}.png")
    graph = pgv.AGraph(dot_file.read())

    query = CallGraphQuery(client.scheduler.backend.session)
    query = query.filter_types({"CallNode", "Value"}).filter_execution_ids([exec_id])

    call_node, result = None, None
    for record in query.all():
        if isinstance(record, CallNode):
            call_node = record
        elif record.value_parsed == (long_hi, True):
            result = record

    arguments = [str((arg.value, arg.arg_hash)) for arg in call_node.arguments]
    result, call_node = str((result, [call_node])), str(call_node)

    expected_nodes = {call_node, result}
    expected_nodes.update(arguments)

    # If the nodes match, adding the graph nodes should have no effect
    produced_nodes = graph.nodes()
    expected_nodes.update(produced_nodes)
    assert len(expected_nodes) == 4

    # Checking truncation
    for node in produced_nodes:
        assert len(node.attr["label"]) <= MAX_LABEL_LEN + len(SEPARATOR)

    produced_edges = graph.edges()
    # Checking twice because there may be duplicate edges that disappear when added to the set
    assert len(produced_edges) == 3
    expected_edges = {(call_node, result), (arguments[0], call_node), (arguments[1], call_node)}
    expected_edges.update(produced_edges)
    assert len(expected_edges) == 3

    # Checking direction
    assert graph.graph_attr["rankdir"] == "TB"

    file.remove()
    dot_file.remove()
    image.remove()


def test_viz_flags():
    x, y = "hi" * 100, "bye" * 100
    result_value = x + y
    argv = ["redun", "run", "workflow.py", "main", "--x", x, "--y", y]
    client, exec_id, file = run_workflow(argv=argv)

    viz_argv = [
        "redun",
        "viz",
        "--hash",
        "--format=dot",
        "--output=test",
        "--horizontal",
        "--jobs",
        "--no-truncation",
        exec_id,
    ]
    client.execute(viz_argv)

    # Testing that the --format and --output specifiers work
    assert not os.path.exists("test.png")
    dot_file = File("test.dot")
    graph = pgv.AGraph(dot_file.read())

    # Checking --horizontal
    assert graph.graph_attr["rankdir"] == "LR"

    query = CallGraphQuery(client.scheduler.backend.session)
    query = query.filter_types({"CallNode", "Value", "Job"}).filter_execution_ids([exec_id])

    call_nodes, jobs, result = [], [], None
    expected_edges = set()
    for record in query.all():
        if isinstance(record, CallNode):
            call_nodes.append(record)
        elif isinstance(record, Job):
            job = str(record)
            jobs.append(job)
            expected_edges.add((job, str(record.call_node)))
            for child_job in record.child_jobs:
                expected_edges.add((job, str(child_job)))
        elif record.value_parsed == result_value:
            result = record

    # Checking --jobs
    assert len(jobs) == 2

    expected_nodes = set(jobs)

    for call_node in call_nodes:
        call_node_str = str(call_node)
        for arg in call_node.arguments:
            arg_str = str((arg.value, arg.arg_hash))
            expected_nodes.add(arg_str)
            expected_edges.add((arg_str, call_node_str))

        for child in call_node.children:
            expected_edges.add((call_node_str, str(child)))

        result_node = str((result, [call_node]))
        expected_nodes.add(result_node)
        expected_nodes.add(call_node_str)
        expected_edges.add((call_node_str, result_node))

    produced_nodes = graph.nodes()
    expected_nodes.update(produced_nodes)
    assert len(expected_nodes) == 10

    produced_edges = graph.edges()
    assert len(expected_edges) == 10
    expected_edges.update(produced_edges)
    assert len(expected_edges) == 10

    # Checking --no-truncation and --hash
    for node in produced_nodes:
        label = node.attr["label"]
        if "Value" in label:
            assert len(label) > MAX_LABEL_LEN
            assert len(label.splitlines()[0].split(" ")) == 2

    file.remove()
    dot_file.remove()


def test_deduplication():
    argv = [
        "redun",
        "run",
        "workflow.py",
        "main",
    ]
    client, exec_id, file = run_workflow(argv)

    result_val = "hibye"
    viz_argv = ["redun", "viz", exec_id, "--format=dot", "--deduplicate"]
    client.execute(viz_argv)

    query = CallGraphQuery(client.scheduler.backend.session)
    query = query.filter_types({"CallNode", "Value"}).filter_execution_ids([exec_id])

    expected_nodes, expected_edges = set(), set()
    call_nodes, result, args = [], None, []
    for record in query.all():
        if isinstance(record, CallNode):
            call_nodes.append(record)
        elif record.value_parsed == result_val:
            result = str((record, None))
        else:
            args.append(str((record, None)))

    for call_node in call_nodes:
        call_node_str = str(call_node)
        for child in call_node.children:
            expected_edges.add((call_node_str, str(child)))

        for arg in args:
            expected_edges.add((arg, call_node_str))

        expected_edges.add((call_node_str, result))

        expected_nodes.add(call_node_str)

    expected_nodes.update(args)
    expected_nodes.add(result)

    dot_file = File(f"{exec_id}.dot")
    graph = pgv.AGraph(dot_file.read())

    expected_nodes.update(graph.nodes())
    assert len(expected_nodes) == 5

    produces_edges = graph.edges()
    assert len(produces_edges) == 7
    expected_edges.update()
    assert len(expected_edges) == 7

    file.remove()
    dot_file.remove()


def test_dataflow():
    file_contents = """
from redun import task, File

@task()
def task1(x):
    return [x["first"], x["second"]] + x["list"]

@task()
def main(x=[File("hi"), File("bye")]):
    return task1({"list": x, "first": x[0], "second": x[1]})
"""

    input = [File("hi"), File("bye")]
    result = input + input
    argv = [
        "redun",
        "run",
        "workflow.py",
        "main",
    ]
    client, exec_id, file = run_workflow(argv, file_contents)

    # Running exec WITHOUT dataflow
    viz_argv = ["redun", "viz", "--format=dot", exec_id]

    client.execute(viz_argv)

    query = CallGraphQuery(client.scheduler.backend.session)
    query = query.filter_types({"CallNode", "Value"}).filter_execution_ids([exec_id])

    expected_nodes, expected_edges = set(), set()
    call_nodes, result, args = [], None, []
    for record in query.all():
        if isinstance(record, CallNode):
            call_nodes.append(record)

    for call_node in call_nodes:
        call_node_str = str(call_node)

        for child in call_node.children:
            expected_edges.add((call_node_str, str(child)))

        for arg in call_node.arguments:
            upstream = arg.upstream if arg.upstream else arg.arg_hash
            arg_str = str((arg.value, upstream))
            expected_nodes.add(arg_str)
            expected_edges.add((arg_str, call_node_str))

        result = str((call_node.value, [call_node]))
        expected_nodes.add(result)
        expected_edges.add((call_node_str, result))

        expected_nodes.add(call_node_str)

    expected_nodes.update(args)
    expected_nodes.add(result)

    dot_file = File(f"{exec_id}.dot")
    graph = pgv.AGraph(dot_file.read())

    expected_nodes.update(graph.nodes())
    assert len(expected_nodes) == 6

    produced_edges = graph.edges()
    assert len(produced_edges) == 5
    expected_edges.update(produced_edges)
    assert len(expected_edges) == 5

    # Running exec WITH dataflow
    viz_argv = ["redun", "viz", "--dataflow", "--format=dot", exec_id]

    client.execute(viz_argv)

    expected_nodes, expected_edges = set(), set()
    call_nodes, values = [], []
    for record in query.all():
        if isinstance(record, CallNode):
            call_nodes.append(record)
        elif isinstance(record, Value) and not record.children:
            values.append(record)

    for call_node in call_nodes:
        call_node_str = str(call_node)

        for child in call_node.children:
            expected_edges.add((call_node_str, str(child)))

        for arg in call_node.arguments:
            for arg_child in arg.value.children:
                arg_str = str((arg_child, arg.arg_hash))
                expected_nodes.add(arg_str)
                expected_edges.add((arg_str, call_node_str))

        for result in values:
            result_str = str((result, [call_node]))

            expected_nodes.add(result_str)
            expected_edges.add((call_node_str, result_str))

        expected_nodes.add(call_node_str)

    dot_file = File(f"{exec_id}.dot")
    graph = pgv.AGraph(dot_file.read())

    expected_nodes.update(graph.nodes())
    assert len(expected_nodes) == 10

    produced_edges = graph.edges()
    assert len(produced_edges) == 9
    expected_edges.update(produced_edges)
    assert len(expected_edges) == 9

    file.remove()
    dot_file.remove()


def extract_edge_hashes(edges):
    edge_hashes = []
    for edge in edges:

        def extract_hash(node):
            return node.partition("=")[2].partition(",")[0].strip("'")

        hashed_edge = (extract_hash(str(edge[0])), extract_hash(str(edge[1])))

        if edge.attr["color"] == redun.visualize.SUBVALUE_EDGE_COLOR:
            hashed_edge = (hashed_edge[0], hashed_edge[1], True)

        edge_hashes.append(hashed_edge)

    return edge_hashes


def extract_subgraph_labels(graph):
    labels = [graph.graph_attr["label"]]
    for subgraph in graph.subgraphs():
        labels.extend(extract_subgraph_labels(subgraph))
    return labels


def test_value_viz():
    file_contents = """
from redun import task, File

@task()
def task3(x):
    return [x["first"], x["second"]]

@task()
def task2(x):
    return task3(x) + x["list"]

@task()
def task1(x):
    return task2(x)

@task()
def task4(x):
    return (x[0], x[1])

@task()
def main(x=[File("hi"), File("bye")]):
    z = task4(x)
    y = task1({"list": x, "first": z[0], "second": z[1]})
    return y[0]
"""

    argv = [
        "redun",
        "run",
        "workflow.py",
        "main",
    ]
    client, exec_id, file = run_workflow(argv, file_contents)
    client.execute(argv)

    query = CallGraphQuery(client.scheduler.backend.session)
    query = query.filter_types({"Execution"}).filter_execution_ids([exec_id])
    exec = query.one()
    value_id = exec.call_node.value.value_hash

    # Testing value visualization WITHOUT wrapping routing calls
    viz_argv = [
        "redun",
        "viz",
        value_id,
        "--format=dot",
    ]

    client.execute(viz_argv)

    expected_edge_hashes = {
        ("489d121b", "95183db3"),
        ("59e143c8", "489d121b"),
        ("59e143c8", "259b3e16"),
        ("59e143c8", "a7e1236b"),
        ("9338943f", "59e143c8", True),
        ("0094c848", "9338943f"),
        ("1cdefdf8", "0094c848"),
        ("1cdefdf8", "6b8f3a12"),
    }

    dot_file = File(f"{value_id}.dot")
    graph = pgv.AGraph(dot_file.read())

    produced_nodes = graph.nodes()
    assert len(produced_nodes) == 9

    produced_edge_hashes = extract_edge_hashes(graph.edges())
    assert len(produced_edge_hashes) == 8
    expected_edge_hashes.update(produced_edge_hashes)
    assert len(expected_edge_hashes) == 9

    viz_argv = [
        "redun",
        "viz",
        value_id,
        "--wrap-calls",
        "--format=dot",
    ]

    client.execute(viz_argv)

    expected_edge_hashes = {
        ("489d121b", "95183db3"),
        ("59e143c8", "489d121b"),
        ("9338943f", "59e143c8", True),
        ("0094c848", "9338943f"),
        ("1cdefdf8", "0094c848"),
    }

    graph = pgv.AGraph(dot_file.read())
    produced_edge_hashes = extract_edge_hashes(graph.edges())
    assert len(produced_edge_hashes) == 5
    expected_edge_hashes.update(produced_edge_hashes)
    assert len(expected_edge_hashes) == 5

    expected_labels = {("CallNode\nmain",), ("CallNode\ntask1", "CallNode\ntask2")}
    for subgraph in graph.subgraphs():
        labels = tuple(extract_subgraph_labels(subgraph))
        expected_labels.add(labels)
    assert len(expected_labels) == 2

    file.remove()
    dot_file.remove()
